
#include <assert.h>
#include <ap_axi_sdata.h>

//#define DB_DEBUG


#define MCR_SIZE 4096
typedef ap_axiu<128,4,5,5> AXI_VAL;

// function prototypes
void standalone_mmult (float A[64][64], float B[64][64], float C[64][64]);
void HLS_accel (AXI_VAL in_stream[2*MCR_SIZE/4], AXI_VAL out_stream[MCR_SIZE/4]);



/* ****************************** C++ TEMPLATES ************************************** */

// reference function
template <typename T, int DIM>
void matrix_multiply_ref(T a[DIM][DIM], T b[DIM][DIM], T out[DIM][DIM])
{

	// matrix multiplication of a A*B matrix
	for (int ia = 0; ia < DIM; ++ia)
		for (int ib = 0; ib < DIM; ++ib)
		{

			float sum = 0;

			for (int id = 0; id < DIM; ++id)

				sum += a[ia][id] * b[id][ib];

			out[ia][ib] = sum;
		}

		return;
}

// --------------------------------------------------------
// function to be accelerated in HW
template <typename T, int DIM>
void mmult_hw(T a[DIM][DIM], T b[DIM][DIM], T out[DIM][DIM])
{

 	int const FACTOR = DIM;
	#pragma HLS INLINE
	//#pragma HLS array_partition variable=a block factor=FACTOR dim=2
	//#pragma HLS array_partition variable=b block factor=FACTOR dim=1

	// matrix multiplication of a A*B matrix
	L1:for (int ia = 0; ia < DIM; ++ia)
		L2:for (int ib = 0; ib < DIM; ++ib)
		{
			//#pragma HLS PIPELINE II=1
			T sum = 0;
			L3:for (int id = 0; id < DIM; ++id)
				sum += a[ia][id] * b[id][ib];
			out[ia][ib] = sum;
		}

		return;
}

// --------------------------------------------------------
// functions to insert and extract elements from an axi stream
// includes conversion to correct data type

template <typename T, int U, int TI, int TD>
void pop_stream(T *dst, ap_axiu <sizeof(T)*8*4,U,TI,TD> const &e)
{
#pragma HLS INLINE

	assert(sizeof(T) == sizeof(int));
	union
	{
		int ival;
		T oval;
	} converter;
	converter.ival = e.data( 31,  0);
    *dst = converter.oval;
	converter.ival = e.data( 63, 32);
    *(dst+1) = converter.oval;
	converter.ival = e.data( 95, 64);
    *(dst+2) = converter.oval;
	converter.ival = e.data(127, 96);
    *(dst+3) = converter.oval;

	volatile ap_uint<sizeof(T)> strb = e.strb;
	volatile ap_uint<sizeof(T)> keep = e.keep;
	volatile ap_uint<U> user = e.user;
	volatile ap_uint<1> last = e.last;
	volatile ap_uint<TI> id = e.id;
	volatile ap_uint<TD> dest = e.dest;

	//return ret;
}

template <typename T, int U, int TI, int TD> ap_axiu <sizeof(T)*8*4,U,TI,TD> push_stream(T const &v1, T const &v2, T const &v3, T const &v4, bool last = false)
{
#pragma HLS INLINE
	ap_axiu<sizeof(T)*8*4,U,TI,TD> e;

	assert(sizeof(T) == sizeof(int));
	union
	{
		int oval;
		T ival;
	} converter;
	converter.ival = v1;
	e.data( 31,  0) = converter.oval;
	converter.ival = v2;
    e.data( 63, 32) = converter.oval;
	converter.ival = v3;
    e.data( 95, 64) = converter.oval;
	converter.ival = v4;
    e.data(127, 96) = converter.oval;

	// set it to sizeof(T) ones
	e.strb = -1;
	e.keep = -1; //e.strb;
	e.user = 0;
	e.last = last ? 1 : 0;
	e.id = 0;
	e.dest = 0;
	return e;
}

// --------------------------------------------------------------------
// function to be accelerated in HW wrapped with AXI4-Stream interface

template <typename T, int DIM, int SIZE, int U, int TI, int TD>
void wrapped_mmult_hw (
	AXI_VAL in_stream[2*SIZE/4],
	AXI_VAL out_stream[SIZE/4])
{

#pragma HLS INLINE

	T a[DIM][DIM];
	T b[DIM][DIM];
	T out[DIM][DIM];

	assert(sizeof(T)*8 == 32);

	// stream in first matrix
	for(int i=0; i<DIM; i++)
		for(int j=0; j<DIM/4; j++)
		{
#pragma HLS PIPELINE II=1
			int k = i*DIM/4+j;
            pop_stream<T,U,TI,TD>(&a[i][j*4], in_stream[k]);
			//a[i][j] = pop_stream<T,U,TI,TD>(in_stream[k]);

		}

		// stream in second matrix
		for(int i=0; i<DIM; i++)
			for(int j=0; j<DIM/4; j++)
			{
#pragma HLS PIPELINE II=1
				int k = i*DIM/4+j+SIZE/4;
				pop_stream<T,U,TI,TD>(&b[i][j*4], in_stream[k]);
				//b[i][j] = pop_stream<T,U,TI,TD>(in_stream[k]);
			}

			// do HW multiplication
			mmult_hw<T, DIM>(a,b,out);

			// stream out result matrix
			for(int i=0; i<DIM; i++)
				for(int j=0; j<DIM/4; j++)
				{
					#pragma HLS PIPELINE II=1
					int k = i*DIM/4+j;
					out_stream[k] = push_stream<T,U,TI,TD>(out[i][j*4], out[i][j*4+1], out[i][j*4+2], out[i][j*4+3], k == (SIZE/4-1));
					//out_stream[k] = push_stream<T,U,TI,TD>(out[i][j],k == (SIZE-1));
				}
				return;

}



// test the functions
template <typename T, int DIM, int SIZE, int U, int TI, int TD>
int test_matrix_mult(void)
{
	int i,j, err;

	T matOp1[DIM][DIM];
	T matOp2[DIM][DIM];
	T matMult_sw[DIM][DIM];
	T matMult_hw[DIM][DIM];

	/** Matrix Initiation */
	for(i = 0; i<DIM; i++)
		for(j = 0; j<DIM; j++)
			matOp1[i][j] = (float)(i+j);

	for(i = 0; i<DIM; i++)
		for(j = 0; j<DIM; j++)
			matOp2[i][j] = (float)(i*j);
	/** End of Initiation */


	printf("DEBUGGING AXI4 STREAMING DATA TYPES!\r\n");

	// prepare data for the DUT
	AXI_VAL inp_stream[2*SIZE/4];
	AXI_VAL out_stream[SIZE/4];

	assert(sizeof(T)*8 == 32);
	// stream in the first input  matrix
	for(int i=0; i<DIM; i++)
		for(int j=0; j<DIM/4; j++)
		{
			int k = i*DIM/4+j;
            inp_stream[k] = push_stream<T,U,TI,TD>(matOp1[i][j*4], matOp1[i][j*4+1], matOp1[i][j*4+2], matOp1[i][j*4+3], 0);
			//inp_stream[k]      = push_stream<T,U,TI,TD>(matOp1[i][j],0);
		}

		// stream in the second input  matrix
		for(int i=0; i<DIM; i++)
			for(int j=0; j<DIM/4; j++)
			{
				int k = i*DIM/4+j;
				inp_stream[k+SIZE/4] = push_stream<T,U,TI,TD>(matOp2[i][j*4], matOp2[i][j*4+1], matOp2[i][j*4+2], matOp2[i][j*4+3], k == (SIZE/4-1));
				//inp_stream[k+SIZE] = push_stream<T,U,TI,TD>(matOp2[i][j],k == (SIZE-1));
			}


			//call the DUT
			wrapped_mmult_hw<T, DIM, SIZE, U, TI, TD>(inp_stream, out_stream);

            /*
            for(int i=0; i< SIZE*2/2; i++)
              printf("%d %x\n", i, inp_stream[i].data);
            for(int i=0; i< SIZE/2; i++)
              printf("%d %x\n", i, out_stream[i].data);
            */
			// extract the output matrix from the out stream 
			for(int i=0; i<DIM; i++)
				for(int j=0; j<DIM/4; j++)
				{
					int k = i*DIM/4+j;
					pop_stream<T,U,TI,TD>(&matMult_hw[i][j*4], out_stream[k]);
					//matMult_hw[i][j] = pop_stream<T,U,TI,TD>(out_stream[k]);
				}

	/* reference Matrix Multiplication */
	matrix_multiply_ref<T, DIM>(matOp1, matOp2, matMult_sw);

	/** Matrix comparison */
	err = 0;
	for (i = 0; (i<DIM && !err); i++)
		for (j = 0; (j<DIM && !err); j++)
          if (matMult_sw[i][j] != matMult_hw[i][j]) {
            printf("NG at %d, %d\n", i, j);
				err++;
          }

	if (err == 0)
		printf("Matrixes identical ... Test successful!\r\n");
	else
		printf("Test failed!\r\n");

	return err;
}


